#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : afc.py
# Contents: agglomerative fuzzy clustering
# Author  : Christian Borgelt
# History : 2016.02.22 file created
#-----------------------------------------------------------------------
from sys  import argv, stderr
from time import time
from math import sqrt, exp

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

def cauchy (d,r):
    d += r
    return 1.0/d if d > 0 else oo

#-----------------------------------------------------------------------

def gauss (d,r):
    return exp(-0.5*d/r)

#-----------------------------------------------------------------------

def dsim_abs (a,b):
    return abs(a-b)

#-----------------------------------------------------------------------

def dsim_sqr (a,b):
    x = a-b; return x*x

#-----------------------------------------------------------------------

def dsim_wgt (a,b):
    return abs(a-b) *(a+b)

#-----------------------------------------------------------------------

def msdegs (disq, pts, cls):
    dsum = dict()               # compute sum of inverse squared dists.
    for p in pts:               # excluding the data point itself
        dsum[p] = sum([disq[p,c] for c in cls])
    msds = dict()               # matrix of membership degrees
    for p in pts:               # traverse the data points
        if dsum[p] >= oo:       # if some distance(s) vanish(es)
            k = len([0 for c in cls if disq[p,c] >= oo])
            for c in cls:       # handle vanishing distances
                msds[p,c] = 1.0/k if disq[p,c] >= oo else 0.0
        else:                   # if no distances vanish
            for c in cls:       # compute the membership degrees
                msds[p,c] = disq[p,c] /dsum[p]
    return msds                 # return the membership degrees

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'agglomerative fuzzy clustering'
    version = 'version 1.2 (2016.02.29)         ' \
            + '(c) 2016   Christian Borgelt'
    fixed   = []                # list of program options
    opts    = {'f': ['mship',  'c'],  # radial function type
               'r': ['radius', 1.0],  # radius for radial function
               'd': ['dissim', 's'],  # cluster dissimilarity mode
               'q': ['quiet',False] } # do not print messages

    if len(argv) <= 1:          # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: htspec.py [options] data')
        print(desc); print(version)
        print('-f#      radial function type                '
                      +'(default: %s)' % opts['f'])
        print('         (c: Cauchy, g: Gaussian function)')
        print('-r#      radius for radial function          '
                      +'(default: %g)' % opts['s'])
        print('-d#      cluster dissimilarity mode          '
                      +'(default: %s)' % opts['d'])
        print('         (s: squared, a: absolute, w: weighted')
        print('-q#      quiet (do not print messages)       '
                      +'(default: %s)' % opts['q'])
        print('data     base name of data files             '
                      +'[required]')
        exit()                  # print usage message and abort

    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(fixed) < 1: fixed.append(a); continue
            else: print('too many fixed arguments'); exit()
        if a[1] not in opts:    # check for an option
            print('unknown option: '+a[:2]); exit()
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    opts = dict([opts[o] for o in opts])

    if not opts['quiet']:       # if to print messages
        stderr.write('afc.py')  # print a startup message
        stderr.write(' - ' +desc +'\n' +version +'\n')
        #stderr.write(opts.items()) # print the options
        #stderr.write(fixed)        # and the fixed arguments

    name = fixed[0] if len(fixed) > 0 else 'iris'
    msfn = opts['mship']        # get the base data name
    drad = opts['radius']       # and the parameters as well as
    dsim = opts['dissim']       # distance and similarity functions
    dist = (lambda d: gauss (d,r=drad)) if msfn == 'g' else \
           (lambda d: cauchy(d,r=drad))
    dsim = dsim_wgt if dsim == 'w' else \
           dsim_sqr if dsim == 's' else dsim_abs

    # --- read data ---
    with open(name+'.dom', 'r') as inp:
        atts = [line.rstrip('\r\n').split(')')[0][4:]
                for line in inp if line[:3] == 'dom']
    with open(name+'.tab', 'r') as inp:
        hdr  = inp.readline().rstrip('\n\r').split()
        data = [line.rstrip('\n\r').split() for line in inp]
    cmap = dict(zip(hdr, range(len(hdr))))
    cids = [cmap[c] for c in atts]
    data = [[float(d[i]) for i in cids] for d in data]

    # for d in data: print(d)     # print the data points

    # --- normalize data ---
    dsum = [0.0 for x in data[0]]
    dsqr = [0.0 for x in data[0]]
    dims = range(len(dsum))     # get the number of dimensions
    for d in data:              # traverse the data points
        for i in dims:          # traverse the dimensions
            dsum[i] += d[i]     # compute sum and sum of squares
            dsqr[i] += d[i]*d[i]
    n = float(len(data))        # get the number of data points
    for i in dims:              # traverse the dimensions
        dsum[i] /= n            # compute mean and standard deviation
        dsqr[i] = sqrt((dsqr[i] -n*dsum[i]*dsum[i]) /n)
        #dsqr[i] = sqrt((dsqr[i] -n*dsum[i]*dsum[i]) /(n-1))
    data = [[(d[i] -dsum[i]) /dsqr[i] for i in dims]
            for d in data]      # z-normalize the data points and
    n    = len(data)            # get the number of data points

    # --- compute raw membership degrees ---
    data.sort()                 # sort them (for duplicate detection)
    disq = dict()               # dictionary for squared distances
    pts  = range(n)             # get the range of cluster indices
    for p in pts:               # traverse the data points
        for c in range(p+1):    # compute the distance matrix
            diff = [a-b for a,b in zip(data[p],data[c])]
            disq[p,c] = disq[c,p] = dist(sum([d*d for d in diff]))
    cls = [0]; mrgs = []        # initialize list of mergers
    for i in pts[1:]:           # find different
        if data[i] != data[i-1]: cls.append(i)
        else:                    mrgs.append((i-1,i,i-1))

    #for m in mrgs: print(m)     # print the executed mergers

    #mx = my = best = -1         # initialize the cluster indices
    #for i in range(len(cls)):   # traverse the cluster pairs and
    #    for k in range(i):      # find pair with smallest distance
    #        x = cls[i]; y = cls[k]
    #        if disq[x,y] > best: best = disq[x,y]; mx = x; my = y
    #c = len(data)               # get the next cluster id and
    #m = (mx,my,c)               # form the merger to execute
    #mrgs.append(m); print(m)    # store and print cluster merger
    #msds = msdegs(disq,pts,cls) # compute membership degrees
    #sx   = sy = 0.0             # init. sum of membership degrees
    #for p in pts:               # traverse the data points
    #    sx += msds[p,mx]        # sum the membership degrees
    #    sy += msds[p,my]        # for the two cluster centers
    #s = sx+sy; sx /= s; sy /= s # compute relative weights
    #ctr  = [sx*x+sy*y for x,y in zip(data[mx],data[my])]
    #data.append(ctr)            # compute new cluster center
    #cls = [i for i in cls if i != mx and i != my]+[c]
    #for p in pts:               # compute distance to new center
    #    diff = [a-b for a,b in zip(data[p],ctr)]
    #    disq[p,c] = disq[c,p] = dist(sum([d*d for d in diff]))

    # --- fuzzy agglomerative clustering ---
    while len(cls) > 1:         # while not all clusters merged
        msds = msdegs(disq,pts,cls) # compute membership degrees
        mx = my = -1; best = oo # initialize the cluster indices
        for x in cls:           # traverse the cluster pairs and
            for y in cls:       # find largest cluster similarity
                if y == x: continue
                m = 0           # init. the cluster similarity
                for p in pts:   # traverse the data points
                    m += dsim(msds[p,x],msds[p,y])
                if m < best: best = m; mx = x; my = y
        c = len(data)           # get the next cluster id and
        m = (mx,my,c)           # form the merger to execute
        mrgs.append(m)          # store and print cluster merger
        # print(m,best)

        sx = sy = 0.0           # init. sum of membership degrees
        for p in pts:           # traverse the data points
            sx += msds[p,mx]    # sum the membership degrees
            sy += msds[p,my]    # for the two cluster centers
        s = sx+sy; sx /= s; sy /= s
        ctr = [sx*x+sy*y for x,y in zip(data[mx],data[my])]

        #ctr = [0.0 for x in data[mx]]; s = 0.0
        #for p in pts:           # traverse the data points
        #    m = msds[p,mx] +msds[p,my]; s += m
        #    for i in dims: ctr[i] += m*data[p][i]
        #ctr = [x/s for x in ctr]

        data.append(ctr)        # compute new cluster center
        cls = [i for i in cls if i != mx and i != my]+[c]
        for p in pts:           # compute distance to new center
            diff = [a-b for a,b in zip(data[p],ctr)]
            disq[p,c] = disq[c,p] = dist(sum([d*d for d in diff]))

    with open(name+'.mp', 'w') as out:
        out.write('u :=  1mm;\n')
        out.write('s := 16mm;\n\n')
        out.write('path p;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (1+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *blue  +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *red   +(2-k) *blue)\n')
        #out.write('  else:         ((k-2) *black +(3-k) *red) fi\n')
        #out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (2+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *black +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *blue  +(2-k) *black)\n')
        #out.write('  elseif k < 3: ((k-2) *red   +(3-k) *blue)\n')
        #out.write('  else:         ((k-3) *black +(4-k) *red) fi\n')
        #out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (2+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *black +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *green +(2-k) *black)\n')
        #out.write('  elseif k < 3: ((k-2) *red   +(3-k) *green)\n')
        #out.write('  else:         ((k-3) *black +(4-k) *red) fi\n')
        #out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (3+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *black +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *blue  +(2-k) *black)\n')
        #out.write('  elseif k < 3: ((k-2) *(1,0.8,0) +(3-k) *blue)\n')
        #out.write('  elseif k < 4: ((k-3) *red   +(4-k) *(1,0.7,0))\n')
        #out.write('  else:         ((k-4) *black +(5-k) *red) fi\n')
        #out.write('enddef;\n\n')

        out.write('vardef hue (expr value, expand) =\n')
        out.write('  save k; numeric k;\n')
        out.write('  k = (3+2expand)*value+(1-expand);\n')
        out.write('  if     k < 1: ((k-0) *0.2white  +(1-k) *black)\n')
        out.write('  elseif k < 2: ((k-1) *0.6green  +(2-k) *0.2white)\n')
        out.write('  elseif k < 3: ((k-2) *(1,0.8,0) +(3-k) *0.6green)\n')
        out.write('  elseif k < 4: ((k-3) *red       +(4-k) *(1,0.7,0))\n')
        out.write('  else:         ((k-4) *black     +(5-k) *red) fi\n')
        out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (4+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: (  0,  0,k+0)\n')
        #out.write('  elseif k < 2: (  0,k-1,  1)\n')
        #out.write('  elseif k < 3: (  0,  1,3-k)\n')
        #out.write('  elseif k < 4: (k-3,  1,  0)\n')
        #out.write('  elseif k < 5: (  1,5-k,  0)\n')
        #out.write('  else:         (6-k,  0,  0) fi\n')
        #out.write('enddef;\n\n')

        out.write('vardef dot (expr x,y,c) =\n')
        out.write('  p := fullcircle scaled u shifted (x,y);\n')
        out.write('  fill p withcolor hue(c,0);\n')
        out.write('  draw p withcolor (0.4*hue(c,0))')
        out.write(        ' withpen pencircle scaled 0.33pt;\n')
        out.write('enddef;\n\n')

        out.write('vardef join (expr a,b,d,c) =\n')
        out.write('  draw a--d--b withcolor hue(c,0)'
                 +' withpen pencircle scaled 0.5pt;\n')
        out.write('  dot(xpart d, ypart d, c);\n')
        out.write('enddef;\n\n')

        fno = 0; a = len(atts)
        for x in range(a-1):    # traverse the attribute pairs
            for y in range(x+1,a):
                out.write('beginfig(%d); %%' % fno)
                out.write(' %s:%s\n' % (atts[x],atts[y]))
                #out.write('  pickup pencircle scaled 0.5pt;\n')
                #out.write('  draw unitsquare scaled 2 shifted (-1,-1)'
                #          ' scaled 2s withcolor 0.7white;\n');
                show = [(d[x],d[y]) for d in data]
                for i in range(n):
                    out.write('  dot(%.5fs,%.5fs,0);\n' % show[i])
                m = len(mrgs)
                for i in range(m):
                    mrg  = mrgs[i]
                    src1 = show[mrg[0]]
                    src2 = show[mrg[1]]
                    dst  = show[mrg[2]]
                    c    = float(i+1)/float(m)
                    out.write('  join((%.5fs,%.5fs),' % src1
                                    +'(%.5fs,%.5fs),' % src2
                                    +'(%.5fs,%.5fs),' % dst
                                    +'%.5f);\n' % c)
                out.write('endfig;\n\n')
                fno += 1        # next figure number

        out.write('end.\n')
